# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import json
import re
import gzip
import zlib
from urllib.parse import quote as url_quote
from urllib.parse import urlparse

from mugen.cookies import DictCookie
from mugen.structures import CaseInsensitiveDict


def default_headers():
    return {
        'User-Agent': 'mugen',
        'Accept': '*/*',
        'Accept-Encoding': 'deflate, gzip',
        'Connection': 'Keep-Alive',
    }


def str_encode(dt, encoding='utf-8'):
    '''
    check dt type, then encoding
    '''

    if isinstance(dt, str):
        return dt.encode(encoding) if encoding else dt
    elif isinstance(dt, bytes):
        return dt
    else:
        raise TypeError('argument must be str or bytes, NOT {!r}'.format(dt))


def form_encode(data):
    '''
    form-encode data
    '''

    assert isinstance(data, dict), 'data must be dict like'

    enc_data = '&'.join(
        ['{}={}'.format(
            k, url_quote(
                v if isinstance(v, str)
                else json.dumps(v, ensure_ascii=False)
                )) for k, v in data.items()
        ]
    )
    return enc_data


def url_params_encode(params):
    if isinstance(params, str):
        return params
    elif isinstance(params, bytes):
        return params
    elif isinstance(params, dict):
        _params = []
        for k, v in params.items():
            _params.append(k + '=' + v)
        return '&'.join(_params)
    else:
        raise TypeError('argument must be str or bytes or dict, NOT {!r}'.format(params))


_re_ip = re.compile(r'^(http://|https://|)\d{1,3}.\d{1,3}.\d{1,3}.\d{1,3}')
def is_ip(netloc):
    return _re_ip.search(netloc) is not None


def parse_headers(lines):
    headers = CaseInsensitiveDict()
    cookies = DictCookie()

    protocol, status_code, ok = lines[0].decode('utf-8').split(' ', 2)

    for line in lines[1:]:
        line = line.decode('utf-8').strip()
        if not line:
            continue

        index = line.find(': ')
        key = line[:index]
        value = line[index + 2:]

        if key.lower() == 'set-cookie':
            cookies.load(value)
            if headers.get(key):
                headers[key] += ', ' + value
        else:
            headers[key] = value

    return (protocol, status_code, ok), headers, cookies


def parse_proxy(proxy_url):
    parser = urlparse(proxy_url)

    proxy_scheme = parser.scheme
    if '@' in parser.netloc:
        user_pwd, host_port = parser.netloc.split('@', 1)
        proxy_host, pt = host_port.split(':', 1)
        proxy_port = int(pt)
        username, password = user_pwd.split(':', 1)
    else:
        proxy_host = parser.netloc.split(':')[0]
        proxy_port = parser.port
        username = None
        password = None
    return proxy_scheme, proxy_host, proxy_port, username, password


def decode_gzip(content):
    assert isinstance(content, bytes)
    return gzip.decompress(content)


def decode_deflate(content):
    assert isinstance(content, bytes)
    try:
        return zlib.decompress(content)
    except Exception:
        return zlib.decompress(content, -zlib.MAX_WBITS)


def find_encoding(content_type):
    if 'charset' in content_type.lower():
        chucks = content_type.split(';')
        for chuck in chucks:
            if 'charset' in chuck.lower():
                cks = chuck.split('=')
                if len(cks) == 1:
                    return None
                else:
                    return cks[-1].strip()



# -*- coding: utf-8 -*-

import json
import re

from http.cookies import (
    BaseCookie,
    Morsel,
    _LegalKeyChars,
)

_LegalValueChars = _LegalKeyChars + ' \[\]'
_CookiePattern = re.compile(r"""
    (?x)                           # This is a verbose pattern
    \s*                            # Optional whitespace at start of cookie
    (?P<key>                       # Start of group 'key'
    [""" + _LegalKeyChars + r"""]+?   # Any word of at least one letter
    )                              # End of group 'key'
    (                              # Optional group: there may not be a value.
    \s*=\s*                          # Equal Sign
    (?P<val>                         # Start of group 'val'
    "(?:[^\\"]|\\.)*"                  # Any doublequoted string
    |                                  # or
    \w{3},\s[\w\d\s-]{9,11}\s[\d:]{8}\sGMT  # Special case for "expires" attr
    |                                  # or
    [""" + _LegalValueChars + r"""]*      # Any word or empty string
    )                                # End of group 'val'
    )?                             # End of optional value group
    \s*                            # Any number of spaces.
    (\s+|;|$)                      # Ending either at space, semicolon, or EOS.
    """, re.ASCII)                 # May be removed if safe.


class DictCookie(BaseCookie):

    def __init__(self, *args, **kwargs):
        super(DictCookie, self).__init__(*args, **kwargs)


    def __repr__(self):
        return '<DictCookie: {}>'.format(json.dumps(self.get_dict(),
                                                    ensure_ascii=False))


    def get_dict(self):
        dictionary = {}
        for key, value in self.items():
            if isinstance(value, Morsel):
                value = value.value

            dictionary[key] = value
        return dictionary


    def format_cookie(self):
        return ' '.join([
            '{}={};'.format(key, value) for key, value in self.get_dict().items()
        ])


    def _BaseCookie__parse_string(self, str, patt=_CookiePattern):
        BaseCookie._BaseCookie__parse_string(self, str, patt=patt)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import logging
import asyncio
import socket
import struct

from urllib.parse import urlparse

import mugen
from mugen.utils import is_ip


class GeneralProxyError(Exception): pass

class SOCKS5AuthError(Exception): pass

class SOCKS5Error(Exception): pass

class ProxyNotPort(Exception): pass


SOCKS4_ERRORS = {
    0x5B: 'Request rejected or failed',
    0x5C: 'Request rejected because SOCKS server cannot connect to identd on the client',
    0x5D: 'Request rejected because the client program and identd report different user-ids'
}

SOCKS5_ERRORS = {
    0x01: 'General SOCKS server failure',
    0x02: 'Connection not allowed by ruleset',
    0x03: 'Network unreachable',
    0x04: 'Host unreachable',
    0x05: 'Connection refused',
    0x06: 'TTL expired',
    0x07: 'Command not supported, or protocol error',
    0x08: 'Address type not supported'
}


@asyncio.coroutine
def get_http_proxy_key(proxy_url, dns_cache):
    urlparser = urlparse(proxy_url)
    # ssl = urlparser.scheme == 'https'
    ssl = False
    host = urlparser.netloc.split(':')[0]
    port = urlparser.port

    if is_ip(host):
        if not port:
            raise ProxyNotPort('proxy: {} has not port'.format(proxy_url))
        key = (host, port, ssl)
    else:
        if not port:
            port = 80
        ip, port = yield from dns_cache.get(host, port)
        key = (ip, port, ssl)
    return key


@asyncio.coroutine
def _make_https_proxy_connection(conn, host, port, recycle=None):
    yield from mugen.request('CONNECT', 'http://{}'.format(host),
                             recycle=recycle, connection=conn)
    yield from conn.ssl_handshake(host)
    return conn


class Socks5Proxy:

    def __init__(self, conn, dest_host, dest_port, ssl, username, password):
        self.conn = conn
        self.dest_host = dest_host
        self.dest_port = dest_port
        self.ssl = ssl
        self.username = username
        self.password = password


    @asyncio.coroutine
    def init(self):
        # 1. connect to socks server and to authorize
        yield from self.auth()

        # 2. let socks server to connect dest_host
        yield from self.connect()

        # 3. SSL/TLS handshake
        if self.ssl:
            yield from self.connect_ssl()

        self.conn.socks_on = True


    @asyncio.coroutine
    def auth(self):
        logging.debug('[Socks5Proxy.init.auth]: {}'.format(self.conn))

        # sending the authentication packages we support.
        if self.username and self.password:
            self.conn.send(b'\x05\x02\x00\x02')
        else:
            self.conn.send(b'\x05\x01\x00')

        chosen_auth = yield from self.conn.read(2)

        if chosen_auth[0:1] != b'\x05':
            raise GeneralProxyError('SOCKS5 proxy server sent invalid data')

        if chosen_auth[1:2] == b'\x02':
            # Okay, we need to perform a basic username/password
            # authentication.
            self.conn.send(b'\x01'
                            + chr(len(self.username)).encode()
                            + self.username
                            + chr(len(self.password)).encode()
                            + self.password)

            auth_status = yield from self.conn.reader.read()
            if auth_status[0:1] != b'\x01':
                # Bad response
                raise GeneralProxyError('SOCKS5 proxy server sent invalid data')
            if auth_status[1:2] != b'\x00':
                # Authentication failed
                raise SOCKS5AuthError('SOCKS5 authentication failed')

        # No authentication is required if 0x00
        elif chosen_auth[1:2] != b'\x00':
            # Reaching here is always bad
            if chosen_auth[1:2] == b'\xFF':
                raise SOCKS5AuthError('All offered SOCKS5 authentication methods were rejected')
            else:
                raise GeneralProxyError('SOCKS5 proxy server sent invalid data')
        # Otherwise, authentication succeeded


    @asyncio.coroutine
    def connect(self):
        logging.debug('[Socks5Proxy.init.connect]: {}'.format(self.conn))

        cmd = b'\x01' # CONNECT
        # Now we can request the actual connection
        header = b'\x05' + cmd + b'\x00'

        family_to_byte = {socket.AF_INET: b'\x01', socket.AF_INET6: b'\x04'}
        for family in (socket.AF_INET, socket.AF_INET6, None):
            if not family:
                self.conn.send(header
                               + b'\x03' + bytes([len(self.dest_host)])
                               + self.dest_host.encode('utf-8')
                               + struct.pack('>H', self.dest_port))
                break

            try:
                addr_bytes = socket.inet_pton(family, self.dest_host)
                self.conn.send(header
                               + family_to_byte[family]
                               + addr_bytes
                               + struct.pack('>H', self.dest_port))
                break
            except socket.error:
                continue

        # Get the response
        resp = yield from self.conn.read(3)
        if resp[0:1] != b'\x05':
            raise GeneralProxyError('SOCKS5 proxy server sent invalid data')

        status = ord(resp[1:2])
        if status != 0x00:
            # Connection failed: server returned an error
            error = SOCKS5_ERRORS.get(status, 'Unknown error')
            raise SOCKS5Error('{0:#04x}: {1}'.format(status, error))

        # Get the bound address/port
        tp = yield from self.conn.read(1)
        if tp == b'\x01':
            chk = yield from self.conn.read(4)
            addr = socket.inet_ntoa(chk)
        elif tp == b'\x03':
            length = yield from self.conn.read(1)
            addr = yield from self.conn.read(ord(length))
        elif tp == b'\x04':
            chk = yield from self.conn.read(16)
            addr = socket.inet_ntop(socket.AF_INET6, chk)
        else:
            raise GeneralProxyError('SOCKS5 proxy server sent invalid data')

        pt = yield from self.conn.read(2)
        port = struct.unpack('>H', pt)[0]
        return addr, port


    @asyncio.coroutine
    def connect_ssl(self):
        logging.debug('[Socks5Proxy.connect_ssl]: {}'.format(self.conn))
        yield from self.conn.ssl_handshake(self.dest_host)
        self.conn.ssl_on = True



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import logging
import asyncio
from urllib.parse import urljoin

from mugen.cookies import DictCookie
from mugen.connection_pool import ConnectionPool
from mugen.connect import Connection
from mugen.adapters import HTTPAdapter
from mugen.utils import is_ip
from mugen.structures import CaseInsensitiveDict
from mugen.models import (
    Request,
    DNSCache,
    DEFAULT_REDIRECT_LIMIT,
    MAX_CONNECTION_POOL,
    MAX_POOL_TASKS,
    MAX_REDIRECTIONS,
    DEFAULT_ENCODING,
)
from mugen.exceptions import (
    RedirectLoop,
    TooManyRedirections
)


class Session(object):

    def __init__(self,
                 headers=None,
                 cookies=None,
                 recycle=True,
                 encoding=None,
                 max_pool=MAX_CONNECTION_POOL,
                 max_tasks=MAX_POOL_TASKS,
                 loop=None):

        logging.debug('instantiate Session: '
                      'max_pool: {}, max_tasks: {}, '
                      'recycle: {}, encoding: {}'.format(
                          max_pool, max_tasks, recycle, encoding))

        self.headers = CaseInsensitiveDict()
        if headers:
            self.headers.update(headers)

        self.cookies = DictCookie()
        if cookies:
            self.cookies = DictCookie.update(cookies)

        self.recycle = recycle
        self.encoding = encoding

        self.max_redirects = DEFAULT_REDIRECT_LIMIT
        self.loop = loop or asyncio.get_event_loop()

        self.connection_pool = ConnectionPool(recycle=recycle,
                                              max_pool=max_pool,
                                              max_tasks=max_tasks,
                                              loop=self.loop)
        self.adapter = HTTPAdapter(self.connection_pool,
                                   recycle=recycle,
                                   loop=self.loop)
        self.dns_cache = DNSCache(loop=self.loop)


    @asyncio.coroutine
    def request(self, method, url,
                params=None,
                headers=None,
                data=None,
                cookies=None,
                proxy=None,
                allow_redirects=True,
                recycle=None,
                encoding=None,
                timeout=None,
                connection=None):

        if recycle is None:
            recycle = self.recycle

        if allow_redirects:
            response = yield from asyncio.wait_for(
                self._redirect(method, url,
                               params=params,
                               headers=headers,
                               data=data,
                               cookies=cookies,
                               proxy=proxy,
                               allow_redirects=allow_redirects,
                               recycle=recycle,
                               encoding=encoding,
                               connection=connection),
                timeout=timeout
            )
        else:
            response = yield from asyncio.wait_for(
                self._request(method, url,
                              params=params,
                              headers=headers,
                              data=data,
                              cookies=cookies,
                              proxy=proxy,
                              allow_redirects=allow_redirects,
                              recycle=recycle,
                              encoding=encoding,
                              connection=connection),
                timeout=timeout
            )

        return response


    @asyncio.coroutine
    def _request(self, method, url,
                 params=None,
                 headers=None,
                 data=None,
                 cookies=None,
                 proxy=None,
                 allow_redirects=True,
                 recycle=None,
                 encoding=None,
                 connection=None):

        logging.debug('[Session.request]: '
                      'method: {}, '
                      'url: {}, '
                      'params: {}, '
                      'headers: {}, '
                      'data: {}, '
                      'cookies: {}, '
                      'proxy: {}'.format(
                          method,
                          url,
                          params,
                          headers,
                          data,
                          cookies,
                          proxy))

        encoding = encoding or self.encoding

        if recycle is None:
            recycle = self.recycle

        if cookies:
            self.cookies.update(cookies)

        if headers is None or not dict(headers):
            headers = self.headers

        request = Request(method, url,
                          params=params,
                          headers=headers,
                          data=data,
                          proxy=proxy,
                          cookies=self.cookies,
                          encoding=encoding)

        # make connection
        if not connection:
            host, *_ = request.url_parse_result.netloc.split(':', 1)
            ssl = request.url_parse_result.scheme.lower() == 'https'
            port = request.url_parse_result.port
            if not port:
                port = 443 if ssl else 80

            if proxy:
                conn = yield from self.adapter.generate_proxy_connect(
                    host, port, ssl, proxy, self.dns_cache, recycle=recycle)
            else:
                conn = yield from self.adapter.generate_direct_connect(
                    host, port, ssl, self.dns_cache, recycle=recycle)
        else:
            if not isinstance(connection, Connection):
                raise TypeError('connection is NOT an instance of Mugen.connect.Connection')

            conn = connection

        # send request
        yield from self.adapter.send_request(conn, request)

        response = yield from self.adapter.get_response(method, conn,
                                                        encoding=encoding)

        # update cookies
        self.cookies.update(response.cookies)
        response.cookies = self.cookies

        if method.lower() != 'connect':
            self.connection_pool.recycle_connection(conn)

        return response


    @asyncio.coroutine
    def _redirect(self, method, url,
                  params=None,
                  headers=None,
                  data=None,
                  cookies=None,
                  proxy=None,
                  allow_redirects=True,
                  recycle=None,
                  encoding=None,
                  connection=None):

        if recycle is None:
            recycle = self.recycle

        history = []
        _URL = url
        base_url = url
        redirect_urls = set()

        while True:
            if len(redirect_urls) > MAX_REDIRECTIONS:
                raise TooManyRedirections(_URL)

            redirect_urls.add(url)
            response = yield from self._request(method, url,
                                                params=params,
                                                headers=headers,
                                                data=data,
                                                cookies=cookies,
                                                proxy=proxy,
                                                allow_redirects=allow_redirects,
                                                recycle=recycle,
                                                encoding=encoding,
                                                connection=connection)

            if not response.headers.get('Location'):
                response.history = history
                return response

            # XXX, not store responses in self.history, which could be used by other
            # coroutines

            location = response.headers['Location']
            url = urljoin(base_url, location)
            base_url = url

            if url in redirect_urls:
                raise RedirectLoop(url)

            history.append(response)


    @asyncio.coroutine
    def head(self, url,
             params=None,
             headers=None,
             cookies=None,
             proxy=None,
             allow_redirects=False,
             recycle=None,
             encoding=None,
             timeout=None,
             connection=None):

        if recycle is None:
            recycle = self.recycle

        response = yield from self.request(
            'HEAD', url,
            params=params,
            headers=headers,
            cookies=cookies,
            proxy=proxy,
            allow_redirects=allow_redirects,
            recycle=recycle,
            encoding=encoding,
            timeout=timeout,
            connection=connection
        )
        return response


    @asyncio.coroutine
    def get(self, url,
            params=None,
            headers=None,
            cookies=None,
            proxy=None,
            allow_redirects=True,
            recycle=None,
            encoding=None,
            timeout=None,
            connection=None):

        if recycle is None:
            recycle = self.recycle

        response = yield from self.request(
            'GET', url,
            params=params,
            headers=headers,
            cookies=cookies,
            proxy=proxy,
            allow_redirects=allow_redirects,
            recycle=recycle,
            encoding=encoding,
            timeout=timeout,
            connection=connection
        )
        return response


    @asyncio.coroutine
    def post(self, url,
             params=None,
             headers=None,
             data=None,
             cookies=None,
             proxy=None,
             allow_redirects=True,
             recycle=None,
             encoding=None,
             timeout=None,
             connection=None):

        if recycle is None:
            recycle = self.recycle

        response = yield from self.request(
            'POST', url,
            params=params,
            headers=headers,
            data=data,
            cookies=cookies,
            proxy=proxy,
            allow_redirects=allow_redirects,
            recycle=recycle,
            encoding=encoding,
            timeout=timeout,
            connection=connection
        )
        return response


    def clear(self):
        """
        Reset cookies and headers to empty
        """

        self.cookies.clear()
        self.headers = None


    def close(self):
        """
        Close this session, all connections and dns cache will be cleaned.
        cookies will be set to None
        """

        # self.adapter.close()   # No sense
        self.connection_pool.clear()
        self.dns_cache.clear()
        self.headers = self.cookies = self.dns_cache = None



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import logging
import asyncio

from collections import defaultdict, deque

from mugen.connect import Connection
from mugen.models import (
    Singleton,
    MAX_CONNECTION_POOL,
    MAX_KEEP_ALIVE_TIME,
    MAX_POOL_TASKS,
    DEFAULT_RECHECK_INTERNAL
)



class ConnectionPool(Singleton):
    '''
    recycle is True, restore connections for reuse
    '''

    def __init__(self,
                 recycle=True,
                 max_pool=MAX_CONNECTION_POOL,
                 max_tasks=MAX_POOL_TASKS,
                 recheck_internal=DEFAULT_RECHECK_INTERNAL,
                 loop=None):

        if hasattr(self, '_initiated'):
            return None

        logging.debug('instantiate ConnectionPool')

        self._initiated = True
        self.recycle = recycle
        self.max_pool = max_pool     # overall pool
        self.max_tasks = max_tasks   # per-key limit
        self.loop = loop or asyncio.get_event_loop()
        self.__connections = defaultdict(deque)
        self.__connection_sizes = defaultdict(int)
        self.__recheck_internal = recheck_internal
        self.__call_count = 0

        asyncio.ensure_future(self._keep_alive_watcher(), loop=loop)


    def __repr__(self):
        return ('<ConnectionPool: ' + 'connections: ' + ', '.join(
            ['{}: {}'.format(key, len(conns))
                for key, conns in self.__connections.items()])
             + ' pool_size: {}'.format(len(self.__connections))
             + '>'
        )


    @asyncio.coroutine
    def _keep_alive_watcher(self):
        # recheck connections for each MAX_KEEP_ALIVE_TIME
        while True:
            yield from asyncio.sleep(MAX_KEEP_ALIVE_TIME)
            try:
                self.recheck_connections()
            except Exception as err:
                logging.error('[ConnectionPool._keep_alive_watcher]: {}'.format(err))


    def get_connections(self, key):
        return self.__connections[key]


    @asyncio.coroutine
    def get_connection(self, key, recycle=None, timeout=None):
        logging.debug(
            '[ConnectionPool.get_connection]: '
            '{}, recycle: {}'.format(key, recycle))

        if recycle is None:
            recycle = self.recycle

        if recycle is False:
            return self.make_connection(key, recycle=recycle, timeout=timeout)

        conns = self.__connections[key]
        while len(conns):
            conn = conns.popleft()
            self.count_connections(key, -1)
            if not conn.stale():
                return conn
            else:
                conn.close()

        if not conns:
            del self.__connections[key]

        conn = self.make_connection(key, recycle=recycle, timeout=timeout)
        return conn


    def make_connection(self, key, recycle=None, timeout=None):
        logging.debug('[ConnectionPool.make_connection]'
                      ': {}, recycle: {}'.format(key, recycle))

        if recycle is None:
            recycle = self.recycle

        ip, port, ssl, *_ = key
        conn = Connection(ip, port,
                          ssl=ssl,
                          key=key,
                          recycle=recycle,
                          timeout=timeout,
                          loop=self.loop)
        return conn


    def recycle_connection(self, conn):
        logging.debug('[ConnectionPool.recycle_connection]: {}'.format(conn))

        if conn.recycle and not conn.stale() and not conn.is_timeout():
            key = conn.key
            conns = self.__connections[key]
            if len(conns) < self.max_tasks or \
                    len(self.__connections) < self.max_pool:
                conns.append(conn)
                self.count_connections(key, 1)
                return None
        conn.close()


    def recheck_connections(self):
        logging.debug('[ConnectionPool.recheck_connections]: {!r}'.format(self))

        empty_conns = []
        for key in self.__connections:
            # to ignore "RuntimeError: dictionary changed size during iteration"
            # when iterating a dictionary
            conns = self.__connections[key]
            conn_num = len(conns)
            for _ in range(conn_num):
                conn = conns.popleft()
                self.count_connections(key, -1)
                self.recycle_connection(conn)
            if not conns:
                empty_conns.append(key)

        for key in empty_conns:
            del self.__connections[key]


    def count_connections(self, key, incr):
        if self.__connection_sizes[key] > 0:
            self.__connection_sizes[key] += incr
        else:
            del self.__connection_sizes[key]


    def clear(self):
        """
        Close all connnections
        """

        logging.debug('[ConnectionPool.clear]')

        for key in self.__connections:
            conns = self.__connections[key]
            while len(conns):
                conn = conns.popleft()
                self.count_connections(key, -1)
                conn.recycle = False
                conn.close()

        self.__connections.clear()


    def closed(self):
        return self._initiated is None and self.__connections is None


    def close(self):
        """
        clear connection_pool and reset the instance to uninitiated
        """

        self.clear()
        self._initiated = self.__connections = None
        self.__connection_sizes = self.loop = None



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

from collections import MutableMapping, OrderedDict


class CaseInsensitiveDict(MutableMapping):

    def __init__(self, *args, **kwargs):
        self._store = OrderedDict()
        self._map = {}

        self.update(*args, **kwargs)

    def __getitem__(self, key):
        if isinstance(key, str):
            key = key.lower()

        return self._store[key]


    def __setitem__(self, key, value):
        if isinstance(key, str):
            lower_key = key.lower()
        else:
            lower_key = key

        self._store[lower_key] = value
        self._map[lower_key] = key


    def __delitem__(self, key):
        if isinstance(key, str):
            key = key.lower()

        self._store.pop(key)
        self._map.pop(key)


    def __iter__(self):
        for k in self._map.values():
            yield k


    def __len__(self):
        return len(self._store)





# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import logging
import asyncio

from mugen.utils import is_ip, parse_proxy
from mugen.exceptions import UnknownProxyScheme
from mugen.proxy import (
    _make_https_proxy_connection,
    Socks5Proxy
)
from mugen.models import (
    Singleton,
    Response,
    DEFAULT_ENCODING
)
# from mugen.connection_pool import ConnectionPool


class HTTPAdapter(Singleton):

    def __init__(self, connection_pool, recycle=True, loop=None):
        if hasattr(self, '_initiated'):
            return None

        logging.debug('instantiate HTTPAdapter: '
                      'recycle: {}, '.format(recycle))

        self._initiated = True
        self.recycle = recycle
        self.loop = loop or asyncio.get_event_loop()
        self.connection_pool = connection_pool


    @asyncio.coroutine
    def generate_direct_connect(self, host, port, ssl, dns_cache, recycle=True):
        key = None
        if is_ip(host):
            ip = host.split(':')[0]
            key = (ip, port, ssl)

        if not key and not ssl:
            ip, port = yield from dns_cache.get(host, port)
            key = (ip, port, ssl)

        if not key and ssl:
            key = (host, port, ssl)

        conn = yield from self.get_connection(key, recycle=recycle)
        return conn


    @asyncio.coroutine
    def generate_proxy_connect(self, host, port, ssl, proxy, dns_cache, recycle=True):
        proxy_scheme, proxy_host, proxy_port, username, password = parse_proxy(proxy)

        proxy_ip, proxy_port = yield from dns_cache.get(proxy_host, proxy_port)
        key = (proxy_ip, proxy_port, False, host)

        if proxy_scheme.lower() == 'http':
            if not ssl:
                key = (proxy_ip, proxy_port, False)   # http proxy not needs CONNECT request
            conn = yield from self.generate_http_proxy_connect(
                key, host, port, ssl, username, password, recycle=recycle)

        elif proxy_scheme.lower() == 'socks5':
            conn = yield from self.generate_socks5_proxy_connect(
                key, host, port, ssl, username, password, recycle=recycle)

        else:
            raise UnknownProxyScheme(proxy_scheme)
        return conn


    @asyncio.coroutine
    def generate_http_proxy_connect(
        self, key, host, port, ssl, username, password, recycle=True):

        conn = yield from self.get_connection(key, recycle=recycle)

        if ssl and not conn.ssl_on:
            logging.debug('[ssl_handshake]: {}'.format(key))
            yield from _make_https_proxy_connection(
                conn, host, port, recycle=recycle)
            conn.ssl_on = True
        return conn


    def generate_socks5_proxy_connect(
        self, key, host, port, ssl, username, password, recycle=True):

        conn = yield from self.get_connection(key, recycle=recycle)
        if conn.socks_on:
            return conn

        socks5_proxy = Socks5Proxy(conn, host, port, ssl, username, password)
        yield from socks5_proxy.init()
        return conn


    @asyncio.coroutine
    def get_connection(self, key, recycle=True):
        conn = yield from self.connection_pool.get_connection(key, recycle=recycle)
        if not conn.reader:
            yield from conn.connect()
        return conn


    @asyncio.coroutine
    def send_request(self, conn, request):
        request_line, headers, data = request.make_request()
        request_line = request_line.encode('utf-8')
        headers = headers.encode('utf-8')
        if isinstance(data, str):
            data = data.encode('utf-8')

        conn.send(request_line + b'\r\n')
        conn.send(headers + b'\r\n')
        conn.send(b'\r\n')
        if data:
            conn.send(data)


    @asyncio.coroutine
    def get_response(self, method, conn, encoding=DEFAULT_ENCODING):
        response = Response(method, conn, encoding=encoding)
        yield from response.receive()

        if response.headers.get('connection') == 'close':
            conn.recycle = False
            conn.close()
        return response


    def closed(self):
        return self._initiated is None and self.connection_pool is None


    def close(self):
        self._initiated = self.connection_pool = self.loop = None



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import time
import logging
import asyncio

from functools import wraps

from mugen.exceptions import ConnectionIsStale
from mugen.models import MAX_CONNECTION_TIMEOUT, MAX_KEEP_ALIVE_TIME


def async_error_proof(gen):
    @wraps(gen)
    @asyncio.coroutine
    def wrap(self, *args, **kwargs):
        try:
            rs = yield from gen(self, *args, **kwargs)
            return rs
        except Exception as err:
            logging.error('[{}]: {}'.format(gen, repr(err)))
            self.close()
            raise err
    return wrap


def error_proof(func):
    @wraps(func)
    def wrap(self, *args, **kwargs):
        try:
            rs = func(self, *args, **kwargs)
            return rs
        except Exception as err:
            logging.error('[{}]: {}'.format(func, repr(err)))
            self.close()
            raise err
    return wrap


class Connection(object):

    def __init__(self, ip, port,
                 ssl=False,
                 key=None,
                 recycle=True,
                 timeout=None,
                 loop=None):

        self.ip = ip
        self.port = port
        self.ssl = ssl
        self.key = key or (ip, port, ssl)
        self.recycle = recycle
        self.loop = loop or asyncio.get_event_loop()
        self.reader = None
        self.writer = None
        self.ssl_on = False  # For http/socks proxy which need ssl connection
        self.socks_on = False  # socks proxy which needs to be initiated
        self.timeout = timeout or MAX_KEEP_ALIVE_TIME
        self.__last_action = time.time()


    def __repr__(self):
        return '<Connection: {!r}>'.format(self.key)


    def _watch(self):
        self.__last_action = time.time()
        return self.__last_action


    def is_timeout(self):
        return time.time() - self.__last_action > self.timeout


    @async_error_proof
    @asyncio.coroutine
    def connect(self):
        logging.debug('[Connection.connect]: {}'.format(self.key))

        reader, writer = yield from asyncio.open_connection(self.ip,
                                                            self.port,
                                                            ssl=self.ssl,
                                                            loop=self.loop)

        self.reader = reader
        self.writer = writer


    @async_error_proof
    @asyncio.coroutine
    def ssl_handshake(self, host):
        logging.debug('[Connection.ssl_handshake]: {}, {}'.format(
            self.key, host))
        transport = self.reader._transport
        raw_socket = transport.get_extra_info('socket', default=None)
        # transport.pause_reading()
        self.reader, self.writer = yield from asyncio.open_connection(
            ssl=True, sock=raw_socket, server_hostname=host)


    @error_proof
    def send(self, data):
        logging.debug('[Connection.send]: {!r}'.format(data))
        self._watch()

        self.writer.write(data)


    @async_error_proof
    @asyncio.coroutine
    def read(self, size=-1):
        logging.debug('[Connection.read]: {}: size = {}'.format(self.key, size))
        self._watch()
        # assert self.closed() is not True, 'connection is closed'
        # assert self.stale() is not True, 'connection is stale'

        if self.stale():
            logging.debug('[Connection.read] [Error] '
                          '[ConnectionIsStale]: {}'.format(self.key))
            raise ConnectionIsStale('{}'.format(self.key))

        if size < 0:
            chuck = yield from asyncio.wait_for(self.reader.read(size),
                                                timeout=MAX_CONNECTION_TIMEOUT)
            return chuck
        else:
            chucks = b''
            while size:
                chuck = yield from asyncio.wait_for(self.reader.read(size),
                                                    timeout=MAX_CONNECTION_TIMEOUT)
                size -= len(chuck)
                chucks += chuck
            return chucks


    @async_error_proof
    @asyncio.coroutine
    def readline(self):
        # assert self.closed() is False, 'connection is closed'
        # assert self.stale() is not True, 'connection is stale'

        if self.stale():
            logging.debug('[Connection.readline] [Error] '
                          '[ConnectionIsStale]: {}'.format(self.key))
            raise ConnectionIsStale('{}'.format(self.key))

        chuck = yield from asyncio.wait_for(self.reader.readline(),
                                            timeout=MAX_CONNECTION_TIMEOUT)

        logging.debug('[Connection.readline]: '
                      '{}: size = {}'.format(self.key, len(chuck)))

        return chuck


    def close(self):
        logging.debug('[Connection.close]: {}, '
                      'recycle: {}'.format(self.key, self.recycle))

        if not self.closed():
            self.reader.feed_eof()
            self.writer.close()
            self.reader = self.writer = None
            logging.debug('[Connection.close]: DONE. {}, '
                          'recycle: {}'.format(self.key, self.recycle))


    def closed(self):
        return self.reader is None or self.writer is None


    def stale(self):
        is_stale = self.reader is None or self.reader.at_eof()
        if is_stale:
            logging.debug('[Connection.stale]: {} is stale'.format(self.key))
        return is_stale




import logging
import asyncio
import mugen

logging.basicConfig(level=logging.DEBUG)


headers = {
'Connection': 'keep-alive',
'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/52.0.2743.116 Safari/537.36',
'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8',
'Accept-Encoding': 'gzip, deflate, sdch',
}


async def task1():
    # resp = await mugen.get('https://httpbin.org/headers', proxy='http://115.226.11.3:9992')
    resp = await mugen.get('https://www.zybang.com/question/a1d38f335ffa88d0ccb14907885db6b1.html', proxy='http://115.226.11.3:9992')
    # resp = await mugen.post('https://httpbin.org/post', data={'a':1, 'b': 2}, proxy='http://115.226.11.3:9992')
    print(list(resp.headers.items()))
    print(resp.text)
    print(len(resp.text))


loop = asyncio.get_event_loop()
tasks = asyncio.wait([task1()])
loop.run_until_complete(tasks)




# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import asyncio

from mugen.session import Session
from mugen.models import MAX_CONNECTION_POOL, MAX_POOL_TASKS


@asyncio.coroutine
def head(url,
         params=None,
         headers=None,
         cookies=None,
         proxy=None,
         allow_redirects=False,
         recycle=True,
         encoding=None,
         timeout=None,
         connection=None,
         loop=None):

    response = yield from request('HEAD', url,
                                   params=params,
                                   headers=headers,
                                   cookies=cookies,
                                   proxy=proxy,
                                   allow_redirects=allow_redirects,
                                   recycle=recycle,
                                   encoding=encoding,
                                   timeout=timeout,
                                   connection=connection,
                                   loop=loop)
    return response


@asyncio.coroutine
def get(url,
        params=None,
        headers=None,
        cookies=None,
        proxy=None,
        allow_redirects=True,
        recycle=True,
        encoding=None,
        timeout=None,
        connection=None,
        loop=None):

    response = yield from request('GET', url,
                                  params=params,
                                  headers=headers,
                                  cookies=cookies,
                                  proxy=proxy,
                                  allow_redirects=allow_redirects,
                                  recycle=recycle,
                                  encoding=encoding,
                                  timeout=timeout,
                                  connection=connection,
                                  loop=loop)
    return response


@asyncio.coroutine
def post(url,
         params=None,
         headers=None,
         data=None,
         cookies=None,
         proxy=None,
         allow_redirects=True,
         recycle=True,
         encoding=None,
         timeout=None,
         connection=None,
         loop=None):

    response = yield from request('POST', url,
                                  params=params,
                                  headers=headers,
                                  data=data,
                                  cookies=cookies,
                                  proxy=proxy,
                                  allow_redirects=allow_redirects,
                                  recycle=recycle,
                                  encoding=encoding,
                                  timeout=timeout,
                                  connection=connection,
                                  loop=loop)
    return response


@asyncio.coroutine
def request(method, url,
            params=None,
            headers=None,
            data=None,
            cookies=None,
            proxy=None,
            allow_redirects=True,
            recycle=True,
            encoding=None,
            timeout=None,
            connection=None,
            loop=None):

    session = Session(recycle=recycle, encoding=encoding, loop=loop)
    response = yield from session.request(method, url,
                                          params=params,
                                          headers=headers,
                                          data=data,
                                          cookies=cookies,
                                          proxy=proxy,
                                          allow_redirects=allow_redirects,
                                          recycle=recycle,
                                          encoding=encoding,
                                          timeout=timeout,
                                          connection=connection)

    return response


def session(headers=None,
            cookies=None,
            recycle=True,
            encoding=None,
            max_pool=MAX_CONNECTION_POOL,
            max_tasks=MAX_POOL_TASKS,
            loop=None):

    return Session(headers=headers,
                   cookies=cookies,
                   recycle=recycle,
                   encoding=encoding,
                   max_pool=max_pool,
                   max_tasks=max_tasks,
                   loop=loop)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import


class NotFindIP(Exception): pass

class RedirectLoop(Exception): pass

class TooManyRedirections(Exception): pass

class ConnectionIsStale(Exception): pass

class UnknownProxyScheme(Exception): pass

class CanNotCreateConnect(Exception): pass




import sys


try:
    from setuptools import setup
except ImportError:
    # Use distutils.core as a fallback.
    # We won't be able to build the Wheel file on Windows.
    from distutils.core import setup

if sys.version_info < (3, 4, 0):
    raise RuntimeError("mugen requires Python 3.4.0+")

version = "0.4.2"

requires = ['httptools']

setup(
    name="mugen",
    version=version,
    author="PeterDing",
    author_email="dfhasyt@gmail.com",
    license="Apache 2.0",

    description="Mugen is library for http asynchronous requests",
    url="http://github.com/PeterDing/mugen",

    install_requires=requires,

    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
    ],

    packages=["mugen"]
)



# -*- coding: utf-8 -*-

import json
import logging
import asyncio
import socket
from urllib.parse import urlparse, ParseResult

from http.cookies import SimpleCookie, Morsel
from collections import OrderedDict

from mugen.cookies import DictCookie
from mugen.exceptions import NotFindIP
from mugen.structures import CaseInsensitiveDict
from mugen.utils import (
    default_headers,
    url_params_encode,
    form_encode,
    decode_gzip,
    decode_deflate,
    find_encoding,
    is_ip
)

from httptools import HttpResponseParser


MAX_CONNECTION_POOL = 100
MAX_POOL_TASKS = 100
MAX_REDIRECTIONS = 1000
MAX_CONNECTION_TIMEOUT = 1 * 60
MAX_KEEP_ALIVE_TIME = 10 * 60
DEFAULT_DNS_CACHE_SIZE = 5000
DEFAULT_REDIRECT_LIMIT = 100
DEFAULT_RECHECK_INTERNAL = 100
HTTP_VERSION = 'HTTP/1.1'
DEFAULT_ENCODING = 'utf-8'


#
# https://magic.io/blog/uvloop-blazing-fast-python-networking/
#
DEFAULT_READ_SIZE = 1024


class Singleton(object):

    def __new__(cls, *args, **kwargs):

        if not hasattr(cls, '_instance'):
            instance = object.__new__(cls)
            cls._instance = instance
        return cls._instance


class Request(object):

    def __init__(self, method, url,
                 params=None,
                 headers=None,
                 data=None,
                 cookies=None,
                 proxy=None,
                 encoding=None):

        self.method = method.upper()
        self.url = url
        self.params = params or {}
        if headers is None:
            headers = {}
        self.headers = CaseInsensitiveDict(headers or default_headers())
        self.data = data
        self.proxy = proxy
        self.encoding = encoding
        if cookies is None:
            self.cookies = DictCookie()
        else:
            self.cookies = cookies

        self.prepare()


    def prepare(self):
        parser = urlparse(self.url)

        scheme = parser.scheme
        host = parser.netloc
        path = parser.path
        _params = parser.params
        query = parser.query
        fragment = parser.fragment

        if self.params:
            enc_params = url_params_encode(self.params)
            query = '{}&{}'.format(query, enc_params)

        self.url_parse_result = ParseResult(
            scheme = scheme or 'http',
            netloc = host,
            path = path,
            params = _params,
            query = query,
            fragment = fragment)

        self.ssl = scheme.lower() == 'https'


    def make_request(self):
        host = self.url_parse_result.netloc
        request_line = self.make_request_line()
        headers = self.make_request_headers(self.method, host, self.headers, self.cookies)
        data = self.make_request_data(self.data)

        # TODO: encoding file

        return request_line, headers, data


    def make_request_line(self):
        method = self.method
        host = self.url_parse_result.netloc
        port = self.url_parse_result.port or 443
        path = self.url_parse_result.path or '/'
        query = self.url_parse_result.query

        if method.lower() == 'connect':
            request_line = '{} {} {}'.format(method,
                                             '{}:{}'.format(host, port),
                                             HTTP_VERSION)
        else:
            uri = path
            if query:
                uri += '?' + query
            request_line = '{} {} {}'.format(method, uri, HTTP_VERSION)
        return request_line


    def make_request_headers(self, method, host, headers, cookies):
        _headers = []

        if not headers.get('host'):
            _headers.append('Host: ' + host + (':443' if method.lower() == 'connect' else ''))

        if method.lower() == 'post' and not self.data:
            _headers.append('Content-Length: 0')

        if self.data:
            data = self.make_request_data(self.data)
            _headers.append('Content-Length: {}'.format(len(data.encode('utf-8'))))
            if isinstance(self.data, dict) and not headers.get('Content-Type'):
                _headers.append('Content-Type: application/x-www-form-urlencoded')

        # add cookies
        if cookies:
            if isinstance(cookies, (DictCookie, SimpleCookie)):
                _cookies = []
                for k in cookies:
                    # TODO, path ?
                    if isinstance(cookies[k], Morsel):
                        v = cookies[k].value
                    else:
                        v = cookies[k]
                    _cookies.append('{}={};'.format(k, v))

                cookie = 'Cookie: ' + ' '.join(_cookies)
                _headers.append(cookie)
            elif isinstance(cookies, dict):
                _cookies = []
                for k, v in cookies.items():
                    _cookies.append('{}={};'.format(k, v))

                cookie = 'Cookie: ' + ' '.join(_cookies)
                _headers.append(cookie)

        # make headers
        for k, v in headers.items():
            _headers.append(k + ': ' + v)
        return '\r\n'.join(_headers)


    def make_request_data(self, data):
        if data is None:
            return data

        enc_data = None
        if isinstance(data, dict):
            enc_data = form_encode(data)
        elif isinstance(data, str):
            enc_data = data
        elif isinstance(data, bytes):
            enc_data = data
        else:
            TypeError('request data must be str or dict, NOT {!r}'.format(data))

        return enc_data


class HttpResonse(object):

    def __init__(self, cookies=None, encoding=None):
        self.headers = CaseInsensitiveDict()
        self.content = b''
        self.encoding = encoding or DEFAULT_ENCODING
        if cookies is None:
            self.cookies = DictCookie()
        else:
            self.cookies = cookies


    # def on_message_begin(self):
        # print('on_message_begin')


    def on_header(self, name, value):
        name = name.decode(self.encoding)
        value = value.decode(self.encoding)
        if name.lower() == 'set-cookie':
            self.cookies.load(value)
            if self.headers.get(name):
                self.headers[name] += ', ' + value
                return None
        self.headers[name] = value


    # def on_headers_complete(self):
        # print(self.headers)


    def on_body(self, value):
        self.content += value


    # def on_message_complete(self):
        # print('on_message_complete')


    # def on_chunk_header(self, headers):
        # print('on_chunk_header')
        # return True


    # def on_chunk_complete(self):
        # print('on_chunk_complete')


class Response(object):

    def __init__(self, method, connection, encoding=None):
        self.method = method
        self.connection = connection
        self.headers = None
        self.content = None
        self.cookies = DictCookie()
        self.encoding = encoding
        self.status_code = None
        self.history = []


    def __repr__(self):
        return '<Response [{}]>'.format(self.status_code)


    @asyncio.coroutine
    def receive(self):
        http_response = HttpResonse(cookies=self.cookies, encoding=self.encoding)
        http_response_parser = HttpResponseParser(http_response)

        conn = self.connection

        # TODO, handle Maximum amount of incoming data to buffer
        chucks = b''
        while True:
            chuck = yield from conn.readline()
            chucks += chuck
            if chuck == b'\r\n':
                break

        http_response_parser.feed_data(chucks)

        self.status_code = http_response_parser.get_status_code()
        headers = http_response.headers
        self.headers = headers

        # (protocol, status_code, ok), headers, cookies = parse_headers(chucks)
        # self.headers = headers
        # self.status_code = status_code
        # self.cookies = cookies

        # TODO, handle redirect

        body = b''
        if self.method.lower() == 'head':    # HEAD
            self.content = body
            return None

        nbytes = headers.get('Content-Length')
        if nbytes:
            nbytes = int(nbytes)
        if nbytes:
            body += yield from conn.read(nbytes)
        else:
            if headers.get('Transfer-Encoding') == 'chunked':
                blocks = []
                while True:
                    size_header = yield from conn.readline()
                    if not size_header:
                        # logging
                        break

                    parts = size_header.split(b';')
                    size = int(parts[0], 16)
                    if size:
                        block = yield from conn.read(size)
                        assert len(block) == size, ('[Response.receive] [Transfer-Encoding]',
                                                    len(block), size)
                        blocks.append(block)

                    crlf = yield from conn.readline()
                    assert crlf == b'\r\n', repr(crlf)
                    if not size:
                        break

                body += b''.join(blocks)
            else:
                # reading until EOF
                pass
                # body += yield from conn.read(-1)

        if body and self.headers.get('Content-Encoding', '').lower() == 'gzip':
            self.content = decode_gzip(body)
        elif body and self.headers.get('Content-Encoding', '').lower() == 'deflate':
            self.content = decode_deflate(body)
        else:
            self.content = body

        if not self.encoding:
            # find charset from content-type
            encoding = find_encoding(self.headers.get('Content-Type', ''))
            if encoding:
                self.encoding = encoding


    @property
    def text(self):
        # TODO, use chardet to detect charset
        encoding = self.encoding or DEFAULT_ENCODING

        return str(self.content, encoding, errors='replace')


    def json(self):
        return json.loads(self.text)


class DNSCache(Singleton):
    """
    DNS Cache
    """

    def __init__(self, size=DEFAULT_DNS_CACHE_SIZE, loop=None):
        if hasattr(self, '_initiated'):
            return None

        logging.debug('instantiate DNSCache: size: {}'.format(size))

        self._initiated = True
        self.__size = size
        self.__hosts = OrderedDict()
        self.loop = loop or asyncio.get_event_loop()


    def __repr__(self):
        return repr(dict(self.__hosts))


    @asyncio.coroutine
    def get(self, host, port, uncache=False):
        if is_ip(host):
            return host, port

        key = (host, port)
        if uncache:
            ipaddrs = yield from self.get_ipaddrs(host, port)
            ipaddr = self.add_host(key, ipaddrs)
        else:
            ipaddr = self.__hosts.get(key)
            if not ipaddr:
                ipaddrs = yield from self.get_ipaddrs(host, port)
                ipaddr = self.add_host(key, ipaddrs)

        self.limit_cache()

        assert ipaddr, NotFindIP(str(key))

        family, type, proto, canonname, (ip, port, *_) = ipaddr
        return ip, port


    @asyncio.coroutine
    def get_ipaddrs(self, host, port):
        ipaddrs = yield from self.loop.getaddrinfo(host, port,
                                                   proto=socket.IPPROTO_TCP)
        return ipaddrs


    def add_host(self, key, ipaddrs):
        for ipaddr in ipaddrs:
            family, type, proto, canonname, (ip, port, *_) = ipaddr
            if (family == socket.AF_INET
                and type == socket.SOCK_STREAM
                and proto == socket.IPPROTO_TCP):
                self.__hosts[key] = ipaddr
                self.__hosts.move_to_end(key, last=False)  # FIFO
                return ipaddr


    def limit_cache(self):
        while len(self.__hosts) > self.__size:
            self.__hosts.popitem()


    def clear(self):
        self.__hosts.clear()
